import hashlib
from dataclasses import dataclass
from enum import IntEnum, IntFlag, auto
from functools import reduce
from struct import Struct
from typing import List, Type, Union

# Can read about the flag in as_command.h (C client)


class Info1Flags(IntFlag):
    empty = 0
    read = auto()
    get_all = auto()
    unused = auto()
    batch_index = auto()
    xdr = auto()
    dont_get_bin_data = auto()
    read_mode_ap_all = auto()
    # Last bit unused


class Info2Flags(IntFlag):
    empty = 0
    write = auto()
    delete = auto()
    generation = auto()
    # apply write if new generation >= old, good for restore
    generation_gt = auto()
    durable_delete = auto()
    create_only = auto()
    unused = auto()
    respond_all_ops = auto()


class Info3Flags(IntFlag):
    empty = 0
    last = auto()
    commit_master = auto()
    unused = auto()
    update_only = auto()
    create_or_replace = auto()
    replace_only = auto()
    sc_read_type = auto()
    sc_read_relax = auto()


class FieldTypes(IntEnum):
    namespace = 0
    setname = 1
    key = 2
    digest = 4
    task_id = 7
    scan_options = 8
    scan_timeout = 9
    scan_rps = 10
    index_range = 22
    index_filter = 23
    index_limit = 24
    index_order = 25
    index_type = 26
    udf_package_name = 30
    udf_function = 31
    udf_arglist = 32
    udf_op = 33
    query_bins = 40
    batch_index = 41
    batch_index_with_set = 42
    predexp = 43


@dataclass
class Field:
    FORMAT = Struct("!IB")
    field_type: FieldTypes
    data: bytes

    def pack(self) -> bytes:
        length = len(self.data) + 1
        return self.FORMAT.pack(length, self.field_type) + self.data

    @classmethod
    def parse(cls: Type["Field"], data: bytes) -> "Field":
        length, field_type = cls.FORMAT.unpack(data[: cls.FORMAT.size])
        data = data[cls.FORMAT.size : length]
        return cls(field_type=field_type, data=data)

    def __len__(self):
        return len(self.data) + self.FORMAT.size


class OperationTypes(IntEnum):
    read = 1
    write = 2
    cdt_read = 3
    cdt_modify = 4
    map_read = 6
    map_modify = 7
    incr = 5
    append = 9
    prepend = 10
    touch = 11
    bit_read = 12
    bit_modify = 13
    delete = 14


class BinType(IntEnum):
    undef = 0
    integer = 1
    double = 2
    string = 3
    blob = 4
    java = 7
    csharp = 8
    python = 9
    ruby = 10
    php = 11
    erlang = 12
    tmap = 19
    tlist = 20
    ldt = 21
    geojson = 23


@dataclass
class BinData:
    data: Union[str, bytes, int, float]
    INTEGER_FORMAT = Struct("!Q")
    DOUBLE_FORMAT = Struct("!d")

    def pack(self) -> bytes:
        if isinstance(self.data, str):
            return self.data.encode("utf-8")
        elif isinstance(self.data, bytes):
            return self.data
        elif isinstance(self.data, int):
            return self.INTEGER_FORMAT.pack(self.data)
        elif isinstance(self.data, float):
            return self.DOUBLE_FORMAT.pack(self.data)
        raise TypeError(
            f"Expected [str, bytes, int, float], received {type(self.data)}"
        )

    @classmethod
    def parse(
        cls: Type["BinData"], data_type: BinType, data: bytes
    ) -> "BinData":
        if data_type == BinType.string:
            return cls(data=data.decode("utf-8"))
        elif data_type == BinType.blob:
            return cls(data=data)
        elif data_type == BinType.integer:
            return cls(data=cls.INTEGER_FORMAT.unpack(data)[0])
        elif data_type == BinType.double:
            return cls(data=cls.DOUBLE_FORMAT.unpack(data)[0])
        raise TypeError(f"Expected BinType value, received {data_type}")

    def __len__(self) -> int:
        if isinstance(self.data, str) or isinstance(self.data, bytes):
            return len(self.data)
        # Equals to integer format size.
        return self.DOUBLE_FORMAT.size


@dataclass
class Bin:
    FORMAT = Struct("BBB")
    btype: BinType
    version: int
    name: str
    data: BinData

    def pack(self) -> bytes:
        base = self.FORMAT.pack(self.btype, self.version, len(self.name))
        return base + self.name.encode("utf-8") + self.data.pack()

    @classmethod
    def parse(cls: Type["Bin"], data: bytes) -> "Bin":
        unpacked = cls.FORMAT.unpack(data[: cls.FORMAT.size])
        btype, version, name_length = unpacked
        name = data[cls.FORMAT.size : cls.FORMAT.size + name_length].decode(
            "utf-8"
        )
        data = data[cls.FORMAT.size + name_length :]
        bin_data = BinData.parse(btype, data)
        return cls(btype=btype, name=name, version=version, data=bin_data,)

    def __len__(self):
        return self.FORMAT.size + len(self.name) + len(self.data)


@dataclass
class Operation:
    # Size, Op, Bin data type, Bin version, Bin name length
    FORMAT = Struct("!IB")
    operation_type: OperationTypes
    data_bin: Bin

    def pack(self) -> bytes:
        length = len(self.data_bin) + 1
        return (
            self.FORMAT.pack(length, self.operation_type) + self.data_bin.pack()
        )

    @classmethod
    def parse(cls: Type["Operation"], data: bytes) -> "Operation":
        unpacked = cls.FORMAT.unpack(data[: cls.FORMAT.size])
        size, operation_type = unpacked
        data_bin = Bin.parse(data[cls.FORMAT.size : cls.FORMAT.size + size])
        return cls(operation_type=operation_type, data_bin=data_bin)

    def __len__(self):
        return len(self.data_bin) + self.FORMAT.size


@dataclass
class Message:
    FORMAT = Struct("!BBBBxBIIIHH")
    info1: Info1Flags
    info2: Info2Flags
    info3: Info3Flags
    transaction_ttl: int
    fields: List[Field]
    operations: List[Operation]
    result_code: int = 0
    generation: int = 0
    record_ttl: int = 0

    def pack(self) -> bytes:
        base = self.FORMAT.pack(
            self.FORMAT.size,
            self.info1,
            self.info2,
            self.info3,
            self.result_code,
            self.generation,
            self.record_ttl,
            self.transaction_ttl,
            len(self.fields),
            len(self.operations),
        )
        fields = reduce(
            lambda x, y: x + y, (field.pack() for field in self.fields), b""
        )
        operations = reduce(
            lambda x, y: x + y, (op.pack() for op in self.operations), b""
        )
        return base + fields + operations

    @classmethod
    def parse(cls: Type["Message"], data: bytes) -> "Message":
        parsed_tuple = cls.FORMAT.unpack(data[: cls.FORMAT.size])
        (
            _size,
            info1,
            info2,
            info3,
            result_code,
            generation,
            ttl,
            transaction_ttl,
            fields_count,
            operations_count,
        ) = parsed_tuple
        data_left = data[cls.FORMAT.size :]
        fields = []
        operations = []
        for _i in range(0, fields_count):
            f = Field.parse(data_left)
            fields.append(f)
            data_left = data[len(f) :]

        for _i in range(0, operations_count):
            op = Operation.parse(data_left)
            operations.append(op)
            data_left = data[len(op) :]

        return cls(
            info1=info1,
            info2=info2,
            info3=info3,
            result_code=result_code,
            generation=generation,
            record_ttl=ttl,
            transaction_ttl=transaction_ttl,
            fields=fields,
            operations=operations,
        )


def digest_key(set_name: bytes, key: str) -> bytes:
    ripe = hashlib.new("ripemd160")
    ripe.update(set_name)
    ripe.update(key.encode("utf-8"))
    return ripe.digest()


def put_key(
    namespace: str, set_name: str, key: str, bin_name: str, value: str
) -> Message:
    set_encoded = set_name.encode("utf-8")
    namespace_field = Field(FieldTypes.namespace, namespace.encode("utf-8"))
    set_field = Field(FieldTypes.setname, set_encoded)

    ripe_digest = digest_key(set_encoded, key)
    key_field = Field(FieldTypes.digest, ripe_digest)

    dbin = BinData(value)
    bin_container = Bin(BinType.string, 0, bin_name, dbin)
    op = Operation(OperationTypes.write, bin_container)
    return Message(
        info1=Info1Flags.empty,
        info2=Info2Flags.write,
        info3=Info3Flags.empty,
        transaction_ttl=1000,
        fields=[namespace_field, set_field, key_field],
        operations=[op],
    )


def get_key(namespace: str, set_name: str, key: str) -> Message:
    set_encoded = set_name.encode("utf-8")
    namespace_field = Field(FieldTypes.namespace, namespace.encode("utf-8"))
    set_field = Field(FieldTypes.setname, set_encoded)

    ripe_digest = digest_key(set_encoded, key)
    key_field = Field(FieldTypes.digest, ripe_digest)

    return Message(
        info1=Info1Flags.read | Info1Flags.get_all,
        info2=Info2Flags.empty,
        info3=Info3Flags.empty,
        transaction_ttl=1000,
        fields=[namespace_field, set_field, key_field],
        operations=[],
    )



from asyncio import StreamReader, StreamWriter, open_connection
from functools import wraps
from typing import Any, Dict, List, Optional

from .protocol.general import AerospikeHeader, AerospikeMessage
from .protocol.message import get_key, put_key


class AerospikeClientNotConnected(Exception):
    pass


def require_connection(func):
    @wraps(func)
    async def wrapper(
        self: "AerospikeClient", *args: List[Any], **kwargs: Dict[Any, Any]
    ) -> Any:
        if not self._reader or not self._writer:
            raise AerospikeClientNotConnected()
        return await func(self, *args, **kwargs)

    return wrapper


class AerospikeClient:

    __slots__ = [
        "host",
        "port",
        "_user",
        "_password",
        "_use_ssl",
        "_reader",
        "_writer",
    ]

    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        use_ssl: bool = False,
        port: int = 3000,
    ):
        self.host: str = host
        self.port: int = port
        self._user: str = user
        self._password: str = password
        self._use_ssl: bool = use_ssl
        self._reader: Optional[StreamReader] = None
        self._writer: Optional[StreamWriter] = None

    async def connect(self):
        self._reader, self._writer = await open_connection(self.host, self.port)

    @require_connection
    async def _get_response(self) -> AerospikeMessage:
        header_data = await self._reader.readexactly(
            AerospikeHeader.FORMAT.sizeof()
        )
        header = AerospikeHeader.parse(header_data)
        message_data = await self._reader.readexactly(header.length)
        message = AerospikeMessage.parse(header_data + message_data)
        return message

    @require_connection
    async def put_key(
        self, namespace: str, set_name: str, key: str, bin_name: str, value: str
    ) -> None:
        message = put_key(namespace, set_name, key, bin_name, value)
        data = AerospikeMessage(message).pack()
        self._writer.write(data)
        await self._writer.drain()
        response = await self._get_response()
        if response.message.result_code != 0:
            raise Exception(
                f"Unexpected result code {response.message.result_code}"
            )

    @require_connection
    async def get_key(self, namespace: str, set_name: str, key: str) -> Any:
        message = get_key(namespace, set_name, key)
        data = AerospikeMessage(message).pack()
        self._writer.write(data)
        await self._writer.drain()
        response = await self._get_response()
        return {
            op.data_bin.name: op.data_bin.data.data
            for op in response.message.operations
        }



from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Dict, Type, Union

from construct import BytesInteger, Const, Container, Int8ub, Struct

from .admin import AdminMessage
from .message import Message


class MessageType(IntEnum):
    info = 1
    admin = 2
    message = 3
    compressed = 4


MESSAGE_TYPE_TO_CLASS: Dict[MessageType, Type[Any]] = {
    MessageType.admin: AdminMessage,
    MessageType.message: Message,
}

MESSAGE_CLASS_TO_TYPE = {
    AdminMessage: MessageType.admin,
    Message: MessageType.message,
}


@dataclass
class AerospikeHeader:
    # Using construct because it's easier for 48bit integer.
    FORMAT = Struct(
        "version" / Const(2, Int8ub),
        "message_type" / Int8ub,
        "length" / BytesInteger(6),
    )
    message_type: MessageType
    length: int

    def pack(self) -> bytes:
        return self.FORMAT.build(
            Container(message_type=self.message_type, length=self.length)
        )

    @classmethod
    def parse(cls: Type["AerospikeHeader"], data: bytes) -> "AerospikeHeader":
        parsed_data = cls.FORMAT.parse(data)
        return cls(
            message_type=parsed_data.message_type, length=parsed_data.length
        )


@dataclass
class AerospikeMessage:

    message: Union[Message]

    def pack(self) -> bytes:
        packed_message = self.message.pack()
        packed_header = AerospikeHeader(
            message_type=MESSAGE_CLASS_TO_TYPE[type(self.message)],
            length=len(packed_message),
        )
        return packed_header.pack() + packed_message

    @classmethod
    def parse(cls: Type["AerospikeMessage"], data: bytes) -> "AerospikeMessage":
        header = AerospikeHeader.parse(data[: AerospikeHeader.FORMAT.sizeof()])
        message_class = MESSAGE_TYPE_TO_CLASS[header.message_type]
        message = message_class.parse(data[AerospikeHeader.FORMAT.sizeof() :])
        return cls(message=message)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['aioaerospike', 'aioaerospike.protocol']

package_data = \
{'': ['*']}

install_requires = \
['bcrypt>=3.1,<4.0', 'construct>=2.9,<3.0']

setup_kwargs = {
    'name': 'aioaerospike',
    'version': '0.1.1',
    'description': 'Async Pure Python implementation of Aerospike client',
    'long_description': '[![codecov](https://codecov.io/gh/aviramha/aioaerospike/branch/master/graph/badge.svg)](https://codecov.io/gh/aviramha/aioaerospike)\n[![Build Status](https://travis-ci.com/aviramha/aioaerospike.svg?branch=master)](https://travis-ci.com/aviramha/aioaerospike)\n\nThis library is planned to be an async API for Aerospike.\nThe library will be Pure-Python, Protocol based on the C Client.\n\n\n** This package is 3rd party, unrelated to Aerospike company **\n\nLicensed under MIT. See LICENSE\n',
    'author': 'Aviram Hassan',
    'author_email': 'aviramyhassan@gmail.com',
    'url': 'https://github.com/aviramha/aioaerospike',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from enum import IntEnum
from struct import Struct
from typing import List, Type

from bcrypt import hashpw

BCRYPT_SALT = b"$2a$10$7EqJtq98hPqEX7fNZaFWoO"


class AdminCommandsType(IntEnum):
    authenticate = 0
    create_user = 1
    drop_user = 2
    set_password = 3
    change_password = 4
    grant_roles = 5
    revoke_roles = 6
    query_users = 9
    create_role = 10
    drop_role = 11
    grant_privileges = 12
    revoke_privileges = 13
    set_whitelist = 14
    query_roles = 16
    login = 20


class FieldTypes(IntEnum):
    user = 0
    password = 1
    old_password = 2
    credential = 3
    clear_password = 4
    session_token = 5
    session_ttl = 6
    roles = 10
    role = 11
    privileges = 12
    whitelist = 13


@dataclass
class Field:
    FORMAT = Struct("!IB")
    field_type: FieldTypes
    data: bytes

    def pack(self) -> bytes:
        length = len(self.data)
        return self.FORMAT.pack(length, self.field_type) + self.data

    @classmethod
    def parse(cls: Type["Field"], data: bytes) -> "Field":
        length, field_type = cls.FORMAT.unpack(data[: cls.FORMAT.size])
        data = data[cls.FORMAT.size : length]
        return cls(field_type=field_type, data=data)

    def __len__(self):
        return len(self.data)


@dataclass
class AdminMessage:
    FORMAT = Struct("!16xBB")
    command_type: AdminCommandsType
    fields: List[Field]

    def pack(self) -> bytes:
        fields_count = len(self.fields)
        fields_data = b""
        for field in self.fields:
            fields_data += field.pack()
        return self.FORMAT.pack(self.command_type, fields_count) + fields_data

    @classmethod
    def parse(cls: Type["AdminMessage"], data: bytes) -> "AdminMessage":
        command_type, fields_count = cls.FORMAT.unpack(data[: cls.FORMAT.size])
        fields = []
        data_left = data[cls.FORMAT.size :]
        for _i in range(fields_count):
            field = Field.parse(data_left)
            fields.append(field)
            data_left = data_left[Field.FORMAT.size + len(field) :]
        return cls(fields=fields, command_type=command_type)

    @classmethod
    def login(cls: Type["AdminMessage"], user: str, password: str) -> bytes:
        hashed_pass = hash_password(password)
        user_field = Field(FieldTypes.user, user.encode("utf-8"))
        password_field = Field(FieldTypes.password, hashed_pass)
        return cls(
            command_type=AdminCommandsType.login,
            fields=[user_field, password_field],
        ).pack()


def hash_password(password: str) -> bytes:
    """
    Hashes password according to Aerospike algorithm
    """
    return hashpw(password.encode("utf-8"), BCRYPT_SALT)
